from __future__ import annotations

from _pytest.cacheprovider import LFPlugin
import pytest

from pytest_lf_skip.constants import Constants
from pytest_lf_skip.lf_skip import LFSkipPlugin

from .hooks import argparse_parse_hook


def pytest_addoption(
    parser: pytest.Parser,
    pluginmanager: pytest.PytestPluginManager,  # noqa: ARG001
) -> None:
    """Add the lf-skip args to pytest."""
    parser.parse = argparse_parse_hook(parser.parse)  # type: ignore[method-assign, assignment]

    parser.addoption(
        *Constants.lf_skip_parser_options,
        action="store_true",
        default=False,
        help=(
            "If --last-failed is enabled, skip tests that have been passed in the last run instead of deselecting them"
        ),
    )

    parser.addoption(
        *Constants.lf_skip_vscode_options,
        action="store_true",
        default=False,
        help="Automatically enable --lf and --lf-skip when pytest is called from vscode.",
    )


@pytest.hookimpl
def pytest_plugin_registered(
    plugin: object,
    manager: pytest.PytestPluginManager,
) -> None:
    """Register the LFSkipPlugin if the LFPlugin is registered."""
    if isinstance(plugin, LFPlugin) and not isinstance(plugin, LFSkipPlugin):
        # add the override hook
        manager.register(LFSkipPlugin(plugin.config), Constants.lf_skip_plugin_name)



from __future__ import annotations

from dataclasses import asdict
import logging
from typing import TYPE_CHECKING

import pytest

from ._helper import replace_file_text

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from pathlib import Path

    from ._types import ExpectedResult

pytest_plugins = "pytester"

logger = logging.getLogger(__name__)


@pytest.fixture
def example_test_file(pytester: pytest.Pytester) -> Path:
    return pytester.makepyfile(
        """
        def test_will_pass():
            assert True

        def test_will_fail():
            assert False
    """,
    )


@pytest.fixture
def factory_test_plugin(  # pragma: no cover (don't need to test the tests)
    pytester: pytest.Pytester,
    example_test_file: Path,
) -> Callable[..., None]:
    """Ensure that lf-skip works."""

    def _test_plugin(
        run_args: Iterable[str],
        expected_results: Iterable[ExpectedResult],
    ) -> None:
        runtime_args: Iterable[str | Path] = (*run_args, example_test_file)

        for index, expected_result in enumerate(expected_results):
            if expected_result.test_file_replace:
                logger.info(
                    "Replacing '%s' with '%s' in test file...",
                    expected_result.test_file_replace.old,
                    expected_result.test_file_replace.new,
                )
                replace_file_text(
                    example_test_file,
                    expected_result.test_file_replace.old,
                    expected_result.test_file_replace.new,
                )

            if expected_result.prepend_args:
                runtime_args = (*expected_result.prepend_args, *runtime_args)

            res = pytester.runpytest(*runtime_args)

            try:
                res.assert_outcomes(**asdict(expected_result.outcome))
            except AssertionError as e:
                msg = f"[{index}] Outcomes didn't match"
                raise AssertionError(msg) from e

            if expected_result.line_searches:
                for line_search in expected_result.line_searches:
                    match_count = 0
                    for line in res.outlines:
                        if line_search.search in line:
                            match_count += 1
                    assert match_count, f"[{index}] '{line_search.search}' didn't appear {line_search.count} times"

    return _test_plugin



from __future__ import annotations

from io import StringIO
import logging
from typing import TYPE_CHECKING

import pytest

from pytest_lf_skip import __version__
from pytest_lf_skip.hooks import logger
from tests._constants import (
    ASSERT_FALSE_TRUE_REPLACE,
    NO_PREVIOUSLY_FAILED_MESSAGE,
    RERUN_PREVIOUS_MESSAGE,
)
from tests._types import ExpectedResult, LineSearch, Outcome

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.mark.parametrize(
    ("run_args", "expected_results"),
    [
        (
            ["--lf"],
            [
                ExpectedResult(
                    Outcome(passed=1, failed=1),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(failed=1, deselected=1),
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(passed=1, deselected=1),
                    test_file_replace=ASSERT_FALSE_TRUE_REPLACE,
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(passed=2),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
            ],
        ),
        (
            ["--lf", "--lf-skip"],
            [
                ExpectedResult(
                    Outcome(passed=1, failed=1),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(
                        skipped=1,
                        failed=1,
                    ),
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(
                        passed=1,
                        skipped=1,
                    ),
                    test_file_replace=ASSERT_FALSE_TRUE_REPLACE,
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(passed=2),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
            ],
        ),
    ],
    ids=("lf", "lf+skip"),
)
def test_lf(
    factory_test_plugin: Callable[..., None],
    run_args: list[str],
    expected_results: list[ExpectedResult],
) -> None:
    factory_test_plugin(
        run_args,
        expected_results,
    )


@pytest.mark.parametrize(
    ("run_args", "expected_results"),
    [
        (
            ["--lf"],
            [
                ExpectedResult(
                    Outcome(passed=1, failed=1),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(failed=1, deselected=1),
                ),
                ExpectedResult(
                    Outcome(passed=1, deselected=1),
                    test_file_replace=ASSERT_FALSE_TRUE_REPLACE,
                    prepend_args=["--lfnf", "none"],
                ),
                ExpectedResult(
                    Outcome(deselected=2),
                    line_searches=(
                        LineSearch(
                            search="run-last-failure: no previously failed tests, deselecting all items.",
                        ),
                    ),
                ),
            ],
        ),
        (
            ["--lf", "--lf-skip"],
            [
                ExpectedResult(
                    Outcome(passed=1, failed=1),
                    line_searches=(
                        LineSearch(
                            search=NO_PREVIOUSLY_FAILED_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(
                        skipped=1,
                        failed=1,
                    ),
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                ),
                ExpectedResult(
                    Outcome(
                        passed=1,
                        skipped=1,
                    ),
                    test_file_replace=ASSERT_FALSE_TRUE_REPLACE,
                    line_searches=(
                        LineSearch(
                            search=RERUN_PREVIOUS_MESSAGE,
                        ),
                    ),
                    prepend_args=["--lfnf", "none"],
                ),
                ExpectedResult(
                    Outcome(skipped=2),
                    line_searches=(
                        LineSearch(
                            search="run-last-failure: no previously failed tests, deselecting all items.",
                        ),
                    ),
                ),
            ],
        ),
    ],
    ids=("all", "all+skip"),
)
def test_lfnf_none(
    factory_test_plugin: Callable[..., None],
    run_args: list[str],
    expected_results: list[ExpectedResult],
) -> None:
    factory_test_plugin(
        run_args,
        expected_results,
    )


def test_package_version() -> None:
    assert all(
        __version__.startswith(initial_version) is False
        for initial_version in [
            "0.0.0",
            "0.0.1",
            "0.1.0",
        ]
    ), f"Version is {__version__} which indicates that git tags have not been cloned"


def test_no_lf_warning(
    pytester: pytest.Pytester,
) -> None:
    with pytest.warns(
        UserWarning,
        match="only works when running with `--lf`.",
    ):
        pytester.runpytest("--lf-skip")


def test_no_vscode_warning(
    pytester: pytest.Pytester,
) -> None:
    with pytest.warns(
        UserWarning,
        match="only works when running from vscode",
    ):
        pytester.runpytest("--auto-lf-skip-vscode")


def test_autoenable_vscode(
    pytester: pytest.Pytester,
) -> None:
    log_stream = StringIO()
    logger.addHandler(logging.StreamHandler(log_stream))

    # NOTE: this is a bit hacky because we're not passing the args in the same way as vscode (with the -p flag)
    # but it works for the test
    pytester.runpytest("--auto-lf-skip-vscode", "vscode_pytest")

    assert "Auto-added --lf and --lf-skip to args" in log_stream.getvalue()



from __future__ import annotations

from tests._types import ExpectedResult, FileReplace, LineSearch, Outcome

RERUN_PREVIOUS_MESSAGE = "run-last-failure: rerun previous 1 failure"
NO_PREVIOUSLY_FAILED_MESSAGE = "run-last-failure: no previously failed tests, not deselecting items."
FIRST_EXPECTED_RESULT = ExpectedResult(
    Outcome(passed=1, failed=1),
    line_searches=(
        LineSearch(
            search=NO_PREVIOUSLY_FAILED_MESSAGE,
        ),
    ),
)
LAST_EXPECTED_RESULT = ExpectedResult(
    Outcome(passed=2),
    line_searches=(
        LineSearch(
            search=NO_PREVIOUSLY_FAILED_MESSAGE,
        ),
    ),
)
ASSERT_FALSE_TRUE_REPLACE = FileReplace(
    old="assert False",
    new="assert True",
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def replace_file_text(path: Path, old: str, new: str) -> None:
    with path.open("r") as f:
        text = f.read()

    text = text.replace(old, new)

    with path.open("w") as f:
        f.write(text)



from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING, Concatenate, ParamSpec, TypeVar, cast
import warnings

from .constants import Constants

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from os import PathLike

    P = ParamSpec("P")
    R = TypeVar("R")

logger = logging.getLogger(__name__)


def argparse_parse_hook(
    func: Callable[Concatenate[Sequence[str | PathLike[str]], P], R],
) -> Callable[Concatenate[Sequence[str | PathLike[str]], P], R]:
    """Add the `--lf` and `--lf-skip` options into the run if pytest has been called by vscode.

    Also warn if the `--lf-skip` options are used without `--lf` being enabled.
    """

    @functools.wraps(func)
    def wrapper(
        args_: list[str | PathLike[str]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        for search_arg in Constants.lf_skip_vscode_options:
            if search_arg in args_:
                if "vscode_pytest" in args_:
                    # vscode_pytest plugin has been passed in, so add our args
                    args_.insert(0, "--lf")
                    args_.insert(0, "--lf-skip")
                    logger.info("Auto-added --lf and --lf-skip to args")
                else:
                    warnings.warn(
                        f"`{search_arg}` only works when running from vscode.",
                        stacklevel=1,
                    )

        for search_arg in Constants.lf_skip_parser_options:
            if search_arg in args_ and all(arg not in ("--lf", "--last-failed") for arg in args_):
                warnings.warn(
                    f"`{search_arg}` only works when running with `--lf`.",
                    stacklevel=1,
                )
                break

        return func(args_, *args, **kwargs)

    return cast("Callable[Concatenate[Sequence[str | PathLike[str]], P], R]", wrapper)



# ruff: noqa: PLR2004 (Pylint: magic-value-comparison)
# /// script
# requires-python = ">=3.9"
# dependencies = [
#     "packaging",
#     "requests",
#     "tomli",
#     "types-requests",
# ]
# ///
from __future__ import annotations

import argparse
from datetime import datetime, timezone
import json
import os
from pathlib import Path
import sys
from typing import TYPE_CHECKING, ClassVar, NoReturn, TypeVar

from packaging.specifiers import InvalidSpecifier, Specifier, SpecifierSet
from packaging.version import (
    parse,
)
import requests
import tomli

if TYPE_CHECKING:
    from packaging.version import Version

    VERSION_TUPLE_TYPE = TypeVar("VERSION_TUPLE_TYPE", bound=tuple[int, ...])


def error(message: str) -> NoReturn:
    """Print an error message to stderr and exit."""
    print(f"Error: {message}", file=sys.stderr)
    sys.exit(1)


class MetaVersionInfo:
    EOL_PYTHON_VERSIONS_URL: ClassVar[str] = "https://endoflife.date/api/python.json"

    _latest_2_version: Version | None = None
    _latest_3_version: Version | None = None

    _latest_eol_version: Version | None = None
    _oldest_supported_version: Version | None = None

    _subversion_min_maxes: dict[tuple[int, int], tuple[Version, Version]] | None = None

    @property
    def latest_2_version(self) -> Version:
        """The latest Python 2 version."""
        if self._latest_2_version is None:
            self._fetch_data()

        if self._latest_2_version is None:
            error("Failed to get the latest Python 2 version.")

        return self._latest_2_version

    @property
    def latest_3_version(self) -> Version:
        """The latest Python 3 version."""
        if self._latest_3_version is None:
            self._fetch_data()

        if self._latest_3_version is None:
            error("Failed to get the latest Python 3 version.")

        return self._latest_3_version

    @property
    def latest_eol_version(self) -> Version:
        """The latest EOL Python version."""
        if self._latest_eol_version is None:
            self._fetch_data()

        if self._latest_eol_version is None:
            error("Failed to get the latest EOL Python version.")

        return self._latest_eol_version

    @property
    def oldest_supported_version(self) -> Version:
        """The oldest supported Python version."""
        if self._oldest_supported_version is None:
            self._fetch_data()

        if self._oldest_supported_version is None:
            error("Failed to get the oldest supported Python version.")

        return self._oldest_supported_version

    @property
    def subversion_min_maxes(self) -> dict[tuple[int, int], tuple[Version, Version]]:
        """Get the latest and earliest releases for each subversion."""
        if self._subversion_min_maxes is None:
            self._fetch_data()

        if self._subversion_min_maxes is None:
            error("Failed to get the subversion min maxes.")

        return self._subversion_min_maxes

    def _fetch_data(self) -> None:
        current_time = datetime.now(timezone.utc)

        response = requests.get(self.EOL_PYTHON_VERSIONS_URL, timeout=5)
        response.raise_for_status()

        data: list[dict[str, str]] = response.json()

        self._subversion_min_maxes = {}

        for item in data:
            version = parse(item["latest"])

            self._subversion_min_maxes[(version.major, version.minor)] = (
                parse(f"{version.major}.{version.minor}"),
                version,
            )

            if version.major == 2:
                self._latest_2_version = (
                    version if self._latest_2_version is None else max(self._latest_2_version, version)
                )
            else:
                self._latest_3_version = (
                    version if self._latest_3_version is None else max(self._latest_3_version, version)
                )

            eol_dt = datetime.fromisoformat(item["eol"]).replace(tzinfo=timezone.utc)
            if eol_dt <= current_time:
                self._latest_eol_version = (
                    version if self._latest_eol_version is None else max(self._latest_eol_version, version)
                )
            else:
                self._oldest_supported_version = (
                    version if self._oldest_supported_version is None else min(self._oldest_supported_version, version)
                )

        if self._oldest_supported_version is not None:
            self._oldest_supported_version = parse(
                f"{self._oldest_supported_version.major}.{self._oldest_supported_version.minor}"
            )


def next_version(version: Version, meta_version_info: MetaVersionInfo) -> Version:
    """Get the next version in the sequence."""
    this_version_release_bounds = meta_version_info.subversion_min_maxes.get((version.major, version.minor), None)
    if this_version_release_bounds is None:
        error(f"Version {version} doesn't exist.")

    min_version, max_version = this_version_release_bounds

    if (next_value := parse(f"{version.major}.{version.minor}.{version.micro + 1}")) > max_version:
        get_next_version = False
        for (subversion_major, subversion_minor), (
            min_version,
            _,
        ) in meta_version_info.subversion_min_maxes.items():
            if get_next_version:
                return min_version
            if version.major == subversion_major and version.minor == subversion_minor:
                get_next_version = True
                continue

    # can't find the actual next version, so just use the guessed next version
    return next_value


def previous_version(version: Version, meta_version_info: MetaVersionInfo) -> Version:
    """Get the previous version in the sequence."""
    this_version_release_bounds = meta_version_info.subversion_min_maxes.get((version.major, version.minor), None)
    if this_version_release_bounds is None:
        if version.major == 4 and version.minor == 0:
            return meta_version_info.latest_3_version
        error(f"Version {version} is not supported for calculation yet.")
    _, max_version = this_version_release_bounds

    if version.micro == 0:
        get_previous_version = False
        for (subversion_major, subversion_minor), (
            _,
            max_version,
        ) in reversed(meta_version_info.subversion_min_maxes.items()):
            if get_previous_version:
                return max_version
            if version.major == subversion_major and version.minor == subversion_minor:
                get_previous_version = True
                continue

    return parse(f"{version.major}.{version.minor}.{version.micro - 1}")


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Get the minimum and maximum supported Python versions from a given pyproject.toml"
    )
    parser.add_argument(
        "pyproject",
        type=Path,
        help="Path to the pyproject.toml file",
    )
    parser.add_argument(
        "--mode",
        type=str,
        choices=["json", "gh-action"],
        default="json",
        help=(
            "The way that the output should be formatted "
            "(gh-action appends to the file in the GITHUB_OUTPUT environment variable)"
        ),
    )

    return parser.parse_args()


def get_requires_python(pyproject_path: Path, meta_version_info: MetaVersionInfo) -> str:
    """Get the requires-python field from the pyproject.toml file."""
    if not pyproject_path.is_file():
        error("The specified path is not a file.")

    with pyproject_path.open("rb") as f:
        data: dict[str, dict[str, str]] = tomli.load(f)
        return data.get("project", {}).get(
            "requires-python",
            f">={next_version(meta_version_info.latest_eol_version, meta_version_info)}",
        )


def parse_specs(specs: str) -> SpecifierSet:
    """Parse the specs string into a SpecifierSet."""
    try:
        return SpecifierSet(specs)
    except InvalidSpecifier as e:
        error(str(e))
        sys.exit(1)


def handle_eq_operator(
    min_version: Version | None, max_version: Version | None, parsed: Version
) -> tuple[Version | None, Version | None]:
    """Handle the == and === operators."""
    return (
        parsed if min_version is None else min(min_version, parsed),
        parsed if max_version is None else max(max_version, parsed),
    )


def handle_gte_operator(min_version: Version | None, parsed: Version) -> Version | None:
    """Handle the >= operator."""
    return parsed if min_version is None else min(min_version, parsed)


def handle_lte_operator(max_version: Version | None, parsed: Version) -> Version | None:
    """Handle the <= operator."""
    return parsed if max_version is None else max(max_version, parsed)


def handle_gt_operator(
    min_version: Version | None, parsed: Version, meta_version_info: MetaVersionInfo
) -> Version | None:
    """Handle the > operator."""
    parsed = next_version(parsed, meta_version_info)
    return parsed if min_version is None else min(min_version, parsed)


def handle_lt_operator(
    max_version: Version | None, parsed: Version, meta_version_info: MetaVersionInfo
) -> Version | None:
    """Handle the < operator."""
    parsed = previous_version(parsed, meta_version_info)
    return parsed if max_version is None else max(max_version, parsed)


def handle_compatible_release_operator(
    min_version: Version | None, max_version: Version | None, parsed: Version, meta_version_info: MetaVersionInfo
) -> tuple[Version | None, Version | None]:
    """Handle the compatible release operator."""
    if len(parsed.release) >= 3:
        min_version = parsed if min_version is None else min(min_version, parsed)
        max_version = (
            next_version(parsed, meta_version_info)
            if max_version is None
            else max(max_version, next_version(parsed, meta_version_info))
        )
    elif len(parsed.release) >= 2:
        min_version = parsed if min_version is None else min(min_version, parsed)
        max_version = (
            parse(f"{parsed.major}.{parsed.minor + 1}")
            if max_version is None
            else max(max_version, parse(f"{parsed.major}.{parsed.minor + 1}"))
        )
    elif len(parsed.release) >= 1:
        min_version = parsed if min_version is None else min(min_version, parsed)
        max_version = (
            parse(f"{parsed.major + 1}") if max_version is None else max(max_version, parse(f"{parsed.major + 1}"))
        )

    return min_version, max_version


def handle_specifier(
    min_version: Version | None, max_version: Version | None, specifier: Specifier, meta_version_info: MetaVersionInfo
) -> tuple[Version | None, Version | None]:
    """Handle a single specifier."""
    version = parse(specifier.version)

    if specifier.operator == ">=":
        min_version = handle_gte_operator(min_version, version)
    elif specifier.operator == "<=":
        max_version = handle_lte_operator(max_version, version)
    elif specifier.operator == ">":
        min_version = handle_gt_operator(min_version, version, meta_version_info)
    elif specifier.operator == "<":
        max_version = handle_lt_operator(max_version, version, meta_version_info)
    elif specifier.operator in {"==", "==="}:
        min_version, max_version = handle_eq_operator(min_version, max_version, version)
    elif specifier.operator == "~=":
        min_version, max_version = handle_compatible_release_operator(
            min_version, max_version, version, meta_version_info
        )

    return min_version, max_version


def calculate_min_max_versions(specifiers: SpecifierSet, meta_version_info: MetaVersionInfo) -> tuple[Version, Version]:
    """Calculate the minimum and maximum supported Python versions."""
    max_version: Version | None = None
    min_version: Version | None = None

    for specifier in specifiers:
        min_version, max_version = handle_specifier(min_version, max_version, specifier, meta_version_info)

    if min_version is None:
        min_version = meta_version_info.oldest_supported_version
    if max_version is None:
        max_version = meta_version_info.latest_3_version

    if min_version > max_version:
        error("The minimum version is greater than the maximum version.")

    return min_version, max_version


def main(args: argparse.Namespace) -> None:
    """Main body of the script."""
    meta_version_info = MetaVersionInfo()

    raw_spec = get_requires_python(args.pyproject, meta_version_info)

    specs = parse_specs(raw_spec)

    min_version, max_version = calculate_min_max_versions(specs, meta_version_info)

    min_version_str, max_version_str = str(min_version), str(max_version)

    if args.mode == "gh-action":
        # GitHub Actions requires the output to be in a specific format
        if not (gh_output_file := os.environ.get("GITHUB_OUTPUT")):
            error("GITHUB_OUTPUT environment variable is not set.")

        gh_output_path = Path(gh_output_file)

        with gh_output_path.open("a") as f:
            f.write(f"min-version={min_version_str}\n")
            f.write(f"max-version={max_version_str}\n")
    else:
        print(
            json.dumps(
                {
                    "min-version": min_version_str,
                    "max-version": max_version_str,
                }
            )
        )


if __name__ == "__main__":
    main(parse_args())



# /// script
# requires-python = ">=3.10"
# dependencies = []
# ///
"""Simple script to validate strings using the conventional commit format.

Used in CI to ensure that PR titles are consistent, allowing for contributions to use any commit style
they want while still allowing us to generate a changelog.
"""

from __future__ import annotations

import argparse
import re
import sys

# The types of changes that are allowed in a message.
ALLOWED_TYPES = ["feat", "fix", "docs", "style", "refactor", "perf", "test", "build", "ci", "chore", "revert"]

# Make sure the message has a scope, e.g. "fix(scope): message"
FORMAT_REGEX = re.compile(r"^([\w ]+)\((\w+)\)\!?: .+")


def validate_commit_message(message: str) -> bool:
    """Validate a commit message against the conventional commit format.

    Args:
        message (str): The commit message to validate.

    Returns:
        bool: True if the message is valid, False otherwise.
    """
    # Check if the message matches the format
    match = FORMAT_REGEX.match(message)
    if not match:
        # If the message doesn't match the format, it is invalid
        print(f"Invalid message format: {message}")
        print("Expected format: <type>(<scope>): <description>")
        return False

    # Check if the type is allowed
    commit_type = match.group(1)
    if commit_type not in ALLOWED_TYPES:
        print(f"Invalid commit type: {commit_type}")
        print(f"Allowed types: {', '.join(ALLOWED_TYPES)}")
        return False

    # Check if the scope is lowercase
    scope = match.group(2)
    if not scope.islower():
        print(f"Invalid scope: {scope}")
        print("Scope must be lowercase.")
        return False

    # If the message is valid, return True
    return True


def parse_args() -> argparse.Namespace:
    """Parse command line arguments.

    Returns:
        argparse.Namespace: The parsed arguments.
    """
    parser = argparse.ArgumentParser(description="Validate commit messages.")
    parser.add_argument("message", type=str, help="The commit message to validate.")
    return parser.parse_args()


def main() -> None:
    """Main function to validate commit messages."""
    args = parse_args()
    message = args.message

    if not validate_commit_message(message):
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations


class Constants:
    lf_plugin_name = "lfplugin"
    lf_collwrapper_plugin_name = "lfplugin-collwrapper"

    lf_skip_plugin_name = "lfskipplugin"
    lf_skip_parser_options = (
        "--last-failed-skip",
        "--lf-skip",
    )

    lf_skip_vscode_options = (
        "--auto-last-failed-skip-vscode",
        "--auto-lf-skip-vscode",
    )


class Config:
    skip_reason = "previously passed"



from __future__ import annotations

from importlib.util import find_spec

import pytest

if not find_spec("scripts.get_supported_py_versions"):
    pytest.skip("get_supported_py_versions script does not exist.", allow_module_level=True)

try:
    from scripts import get_supported_py_versions as _get_supported_py_versions  # noqa: F401
except ImportError as e:
    pytest.skip(f"couldn't import get_supported_py_versions script ({e}).", allow_module_level=True)

from packaging.version import (
    Version,
    parse,
)

from scripts.get_supported_py_versions import (
    MetaVersionInfo,
    calculate_min_max_versions,
    next_version,
    parse_specs,
    previous_version,
)


@pytest.fixture(autouse=True)
def mocked_version_info() -> MetaVersionInfo:
    info = MetaVersionInfo()
    info._latest_2_version = parse("2.7.18")  # noqa: SLF001
    info._latest_3_version = parse("3.13.3")  # noqa: SLF001
    info._latest_eol_version = parse("3.8.20")  # noqa: SLF001
    info._oldest_supported_version = parse("3.9")  # noqa: SLF001
    info._subversion_min_maxes = {  # noqa: SLF001
        (2, 7): (parse("2.7"), parse("2.7.18")),
        (3, 0): (parse("3.0"), parse("3.0.1")),
        (3, 1): (parse("3.1"), parse("3.1.5")),
        (3, 2): (parse("3.2"), parse("3.2.6")),
        (3, 3): (parse("3.3"), parse("3.3.7")),
        (3, 4): (parse("3.4"), parse("3.4.10")),
        (3, 5): (parse("3.5"), parse("3.5.10")),
        (3, 6): (parse("3.6"), parse("3.6.15")),
        (3, 7): (parse("3.7"), parse("3.7.17")),
        (3, 8): (parse("3.8"), parse("3.8.20")),
        (3, 9): (parse("3.9"), parse("3.9.22")),
        (3, 10): (parse("3.10"), parse("3.10.17")),
        (3, 11): (parse("3.11"), parse("3.11.12")),
        (3, 12): (parse("3.12"), parse("3.12.10")),
        (3, 13): (parse("3.13"), parse("3.13.3")),
    }
    return info


@pytest.mark.parametrize(
    ("version", "expected"),
    [
        ("2.7.18", "3.0"),
        ("3.10", "3.10.1"),
        ("3.10.16", "3.10.17"),
        ("3.10.17", "3.11"),
        ("3.13.3", "3.13.4"),
    ],
)
def test_next_version(version: str, expected: str, mocked_version_info: MetaVersionInfo) -> None:
    """Test the next_version function."""
    assert next_version(parse(version), mocked_version_info) == parse(expected)


@pytest.mark.parametrize(
    ("version", "expected"),
    [
        ("2.7.18", "2.7.17"),
        ("3.10", "3.9.22"),
        ("3.10.16", "3.10.15"),
        ("3.10.17", "3.10.16"),
        ("3.13.3", "3.13.2"),
    ],
)
def test_previous_version(version: str, expected: str, mocked_version_info: MetaVersionInfo) -> None:
    """Test the previous_version function."""
    assert previous_version(parse(version), mocked_version_info) == parse(expected)


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (">=3.9, <4.0.0", (parse("3.9"), parse("3.13.3"))),
        (">=3.8, <3.9", (parse("3.8"), parse("3.8.20"))),
        (">=2.7, <3", (parse("2.7"), parse("2.7.18"))),
        (">=3.10, <=3.11", (parse("3.10"), parse("3.11"))),
    ],
)
def test_calculate_min_max_versions(
    spec: str, expected: tuple[Version, Version], mocked_version_info: MetaVersionInfo
) -> None:
    """Test the parse_specs function."""
    min_version, max_version = calculate_min_max_versions(parse_specs(spec), mocked_version_info)
    assert min_version == expected[0]
    assert max_version == expected[1]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from _pytest.cacheprovider import LFPlugin
import pytest

from .constants import Config, Constants

if TYPE_CHECKING:
    from collections.abc import Generator

    import _pytest.nodes


class LFSkipPlugin(LFPlugin):
    _parent: LFPlugin | None = None

    def __init__(self, config: pytest.Config) -> None:
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        should_be_active = True

        if not config.getoption(Constants.lf_skip_parser_options[0]):
            should_be_active = False

        # check if lfplugin is registered, if not, do nothing
        if should_be_active and (parent := config.pluginmanager.get_plugin(name=Constants.lf_plugin_name)):
            self._parent = parent
        else:
            should_be_active = False

        # save the current state of the lf options
        lf_attr = config._opt2dest.get("lf", "lf")  # noqa: SLF001
        existing_lf_value = config.getoption("lf")

        if not existing_lf_value:
            should_be_active = False

        if should_be_active:
            # allow re-setting of lfplugin-collwrapper, because it gets added in the init function
            # the wording of the function seems backwards to me, but it works.
            # this will allow us to reregister the `lfplugin-collwrapper` to this version when we run super's init

            config.pluginmanager.set_blocked(Constants.lf_collwrapper_plugin_name)
        else:
            # toggle the lf cli arg off temporarily so the lfplugin-collwrapper doesn't get reregistered
            setattr(config.option, lf_attr, False)

        super().__init__(config)

        if should_be_active:
            # disable the lfplugin-collwrapper plugin so it doesn't filter out known passes
            config.pluginmanager.unregister(name=Constants.lf_collwrapper_plugin_name)
            self.logger.info("LFSkipPlugin activated - will skip instead of deselect last failed tests")
        else:
            # set the lf arg pack to it's previous state
            setattr(config.option, lf_attr, existing_lf_value)
            self.active = False
            self.logger.debug(
                "LFSkipPlugin not active - either --lf-skip not provided, --lf not provided, or LFPlugin not found"
            )

    # wrap the existing pytest_collection_modifyitems on the parent
    @pytest.hookimpl(wrapper=True, tryfirst=True)
    def pytest_collection_modifyitems(
        self,
        config: pytest.Config,  # noqa: ARG002
        items: list[_pytest.nodes.Item],
    ) -> Generator[None]:
        """Mark deselected tests as skipped."""
        if not self._parent or not self._parent.active:
            self.logger.debug("Parent LFPlugin not active, skipping LFSkipPlugin hook")
            return (yield)

        self.logger.debug("Running LFSkipPlugin collection modification")
        re_add_items = []

        # Temporarily disable the pytest_deselected hook so that stats aren't
        # messed up - very hacky, but it works
        # (If we could nicely check the calling plugin from within a hook, then
        # that would be better, but for now this is the solution)
        original_deselect_hook = self._parent.config.hook.pytest_deselected

        def pytest_deselected_override(items: list[pytest.Item]) -> None:
            self.logger.info("Converting %d deselected tests to skipped", len(items))
            for item in items:
                item.add_marker(pytest.mark.skip(reason=Config.skip_reason))
                re_add_items.append(item)

        self._parent.config.hook.pytest_deselected = pytest_deselected_override  # type: ignore[attr-defined]

        # Let the parent's pytest_collection_modifyitems execute
        res = yield

        # re-enable the pytest_deselected hook
        self._parent.config.hook.pytest_deselected = original_deselect_hook  # type: ignore[attr-defined]

        # add the previously removed tests back
        items.extend(re_add_items)

        # copy attrs from parent so that output is consistent with standard --lf
        for attr_name in dir(self._parent):
            if attr_name.startswith("__"):
                continue

            if callable(attr := getattr(self._parent, attr_name)):
                continue
            setattr(self, attr_name, attr)

        # deactivate parent LFPlugin to prevent doubled messages
        self._parent.active = False

        return res



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass
class Outcome:
    passed: int = 0
    skipped: int = 0
    failed: int = 0
    deselected: int = 0


@dataclass
class LineSearch:
    search: str
    count: int = 1


@dataclass
class FileReplace:
    old: str
    new: str


@dataclass
class ExpectedResult:
    outcome: Outcome
    test_file_replace: FileReplace | None = None
    line_searches: Iterable[LineSearch] | None = None
    prepend_args: Iterable[str] | None = None
